import { Type, constructorOf, Store } from '@tsed/core';
import { GlobalProviders, InjectorService, registerProvider } from '@tsed/di';

import * as Electron from 'electron';

import { AppOptions } from './decorators/app-settings';
import { AppSettingsService } from './services/app-settings.service';
import { ELECTRON_APP } from './types/electron-app';
import {
  AppProviderMetadata,
  AppEventHandlerMetadata
} from './models/app-provider.metadata';
import { createElectronApp } from './utils/electron';
import { ElectronApp } from './decorators/electron-app';

export abstract class AppLoader {
  readonly injector: InjectorService;
  private startedAt = new Date();

  constructor(settings: Partial<AppOptions> = {}) {
    // create injector with initial configuration
    this.injector = this.createInjector(this.getConfiguration(this, settings));

    createElectronApp(this.injector);
    this.attachEventHandler();
  }

  get settings(): AppSettingsService {
    return this.injector.settings as AppSettingsService;
  }

  get electronApp(): Electron.App {
    return this.injector.get<ElectronApp>(ElectronApp);
  }

  static async bootstrap<App extends AppLoader>(
    module: Type<App>,
    settings: Partial<AppOptions> = {}
  ): Promise<App> {
    const app = new module(settings);

    return app;
  }

  async start(): Promise<any> {
    try {
    } catch (err) {
      return Promise.reject(err);
    }
  }

  private createInjector(settings: Partial<AppOptions> = {}) {
    const injector = new InjectorService();
    injector.settings = this.createSettingsService(injector);
    // injector.logger = $log;

    // @ts-ignore
    injector.settings.set(settings);

    /* istanbul ignore next */
    if (injector.settings.env === 'test') {
      injector.logger.stop();
    }

    return injector;
  }

  private createSettingsService(injector: InjectorService): AppSettingsService {
    const provider = GlobalProviders.get(AppSettingsService)!.clone();

    provider.instance = injector.invoke<AppSettingsService>(provider.useClass);
    injector.addProvider(AppSettingsService, provider);

    return provider.instance as any;
  }

  private getConfiguration(module: any, configuration: any = {}) {
    const provider = GlobalProviders.get(constructorOf(module))!;

    return { ...provider.configuration, ...configuration };
  }

  private async invoke(
    instance: any,
    handlerMetadata: AppEventHandlerMetadata,
    args: any[]
  ): Promise<any> {
    const { methodClassName } = handlerMetadata;

    return await instance[methodClassName](...args);
  }

  private attachEventHandler() {
    const handlerMetadata: AppProviderMetadata = Store.from(this).get(
      ELECTRON_APP
    );

    if (!handlerMetadata || !handlerMetadata.handlers) {
      return;
    }

    const __this = this;

    for (const handler in handlerMetadata.handlers) {
      if (handlerMetadata.handlers.hasOwnProperty(handler)) {
        const metadata = handlerMetadata.handlers[handler];

        Electron.app.on(metadata.channel as any, (...args: any[]) => {
          __this.invoke(__this, metadata, args);
        });
      }
    }
  }
}
